__version__ = '0.1.2'

import airfrans.dataset as dataset
import airfrans.naca_generator as naca_generator
import airfrans.sampling as sampling
from airfrans.reorganize import reorganize
from airfrans.simulation import Simulation